from Foundation import *
from AppKit import *
from PyObjCTools import NibClassBuilder, AppHelper

NibClassBuilder.extractClasses("Flame")


class Flame(NibClassBuilder.AutoBaseClass):

    def applicationDidFinishLaunching_(self, aNotification):
        self.outline.setDoubleAction_("outlineDoubleClicked:")
        self.outline.reloadData()

        # all browsers, keyed on type
        self.browsers = {}
        
        # alll services, keyed on 'id', for various bad reasons.
        self.services = {}
        
        # all hosts, in order. This is the list the outline view uses
        # as it's source for data.
        self.hosts = []

        # for all the protocols that we know about (including the meta
        # protocol for service type discovery) start up a Browser for
        # that protocol and delegate it to us. Don't start it up yet,
        # though.
        for p in protocols:
            b = Browser.new()
            b.setDelegate(self)
            b.setType( p['type'] )
            self.browsers[ p['type'] ] = b

        # start a scan of everything.
        self.rescan_(None)

    def newService_(self, type):
        # the callback from the meta-service discovery - we create a
        # new browser and start a search with it immediately.
        try:
            # try to find a browser
            b = self.browsers[ type ]
        except (KeyError):
            b = Browser.new()
            b.setDelegate(self)
            b.setType( type )
            self.browsers[ type ] = b
            b.search()

    def rescan_(self, object):
        self.services = {}
        self.spinny.startAnimation_(self)
        for b in self.browsers:
            self.browsers[b].search()

    def gotMoreServices(self, browser):
        for serviceid in browser.services:
            self.services[ serviceid ] = Service.new().service( browser.services[ serviceid ], browser.type )

        # schedule a rebuild of the list very soon.
        NSObject.cancelPreviousPerformRequestsWithTarget_selector_object_(self, "filter", None)
        self.performSelector_withObject_afterDelay_("filter", None, 0.5)
        
        # schedule a removal of all hosts with no services in
        NSObject.cancelPreviousPerformRequestsWithTarget_selector_object_(self, "removeEmptyHosts", None)
        self.performSelector_withObject_afterDelay_("removeEmptyHosts", None, 3)

    def removeEmptyHosts(self):
        # called 5 seconds after a service is found, this removes any hosts
        # in the list that are still there, but have no services left.
        for h in self.hosts:
            if len(h.services) == 0:
                self.hosts.remove(h)

        # re-display the outline list
        self.outline.reloadData()

    
    def host_sort(self, first, second):
        return cmp(first.name, second.name)

    def filter(self):
        # filter gets called after there have been no service responses
        # for a while, because it's expensive (from gotMoreServices)
        # If we're getting called, it'll be because we have most of the
        # services we expect to see, so stop the spinner. (it'll hide
        # by itself)
        self.spinny.stopAnimation_(self)

        # hosts in the outline view are Host objects. We want to hang onto
        # the same host objects we had last run, so the outline view will
        # persist the open/closed state of the hosts. Services aren't so
        # important, as they don't have state. We key the hosts by IP in a
        # local variable for uniqueness and easy lookup.
        hosts = {}
        for h in self.hosts:
            hosts[ h.ip ] = h
            # wipe the services of the host - we're starting from scratch
            h.services = []

        # self.services is a dictionary of every service we know about,
        # keyed on object ID (a cheap uniqueness trick). For each of these,
        # we want to assign to a host.
        for serviceid in self.services:
            service = self.services[ serviceid ]

            # get the host with the ip of the service.
            ip = service.ip
            try:
                # do we know about this host?
                host = hosts[ ip ]
            except (KeyError):
                # apparently not. Generate a new object.
                host = Host.new()
                hosts[ ip ] = host
                host.ip = ip
                # add to the list of hosts the window displays
                self.hosts.append(host)

            # fix strange behaviour where we see services more than
            # once, but uniquing the service on ip/type/port
            seen = 0
            for s in host.services:
                if s.ip == service.ip and s.port == service.port and s.type == service.type:
                    seen = 1
                    
            if not seen: host.addService(service)

        # because self.hosts is all the hosts that we knew about _before_
        # filter, not a new list, it's possible that someone has dropped off
        # the network, and therefore has no services. Drop hosts with
        # no services from the list
        # TODO - this doesn't work, damnit, we're called too early and we
        # lose all the hosts. We need to keep the objects, but only _display_
        # the populated ones.
        #for h in self.hosts:
        #    if not len(h.services): self.hosts.remove(h)


        # sort the hosts, in this case, alphabetically by their names.
        self.hosts.sort( self.host_sort )

        # re-display the outline list
        self.outline.reloadData()

    # when we come to the foregound
    def applicationDidBecomeActive_(self, notification):
        self.window.makeKeyAndOrderFront_(self)

    # window closing..
    def windowWillClose_(self, window):
        pass

    def outlineView_numberOfChildrenOfItem_(self, outline, item):
        if item == None:
            try: return len(self.hosts)
            except (AttributeError): return 0

        try:
            return len(item.services)
        except (AttributeError):
            return 0

    def outlineView_isItemExpandable_(self, outline, item):
        try:
            len(item.services)
            return 1
        except (AttributeError):
            return 0
    
    def outlineView_child_ofItem_(self, outline, child, item):
        if item == None:
            return self.hosts[child]

        return item.services[child]
        
    def outlineView_objectValueForTableColumn_byItem_(self, outline, column, item):
        try:
            if column.identifier() != 'name':
                cell = column.dataCell()
                color = NSColor.grayColor()
                cell.setTextColor_(color)
            return item.__dict__[ column.identifier() ]
        except (KeyError):
            return None

    def outlineViewSelectionDidChange_(self, notification):
        outline = notification.object()
        if outline.selectedRow() < 0:
            pass
        else:
            item = outline.itemAtRow_( outline.selectedRow() )
            self.selectedService = item
            self.properties.reloadData()

    def outlineDoubleClicked_(self, outline):
        pass

    def numberOfRowsInTableView_(self, table):
        try:
            return len(self.selectedService.__dict__)
        except (AttributeError):
            # startup not complete yet
            return 0
    
    def tableView_objectValueForTableColumn_row_(self, table, column, row):
        col = column.identifier()

        keys = self.selectedService.__dict__.keys()
        keys.sort()
        key = keys[row]
        if col == "property":
            return key
        else:
            return str( self.selectedService.__dict__[key] )


###################################################################
# NSNetServiceBrowser handling code

class Browser(NSObject):

    def init(self):
      self = NSObject.init(self)
      self.delegate = None
      return self
    
    def setDelegate(self, delegate):
        self.delegate = delegate
        
    def setType(self, type):
        self.type = type
        
    def search(self):
        self.services = {}

        self.b = NSNetServiceBrowser.new()
        self.b.setDelegate_(self)
        self.b.searchForServicesOfType_inDomain_(self.type, "")

    def netServiceBrowserWillSearch_(self, browser):
        browser.performSelector_withObject_afterDelay_("stop", None, 500)

    def netServiceBrowser_didNotSearch_(self, browser, reason):
        NSLog("Browser did not search: " + repr(reason) )

    def netServiceBrowserDidStopSearch_(self, browser):
        pass

    def netServiceBrowser_didFindService_moreComing_(self, browser, service, more):
        
        # if we're being told about a type of service, and we have a delegate,
        # tell the delegate about it.
        if service.type() == "_tcp.local." or service.type() == "_udp.local.":
            if self.delegate:
                type = "%s.%s"%( service.name(), service.type() )
                type = type[:-6]
                return self.delegate.newService_( type )


        service.setDelegate_(self)
        service.resolve()

        # leave the browser running, even if there are no more services, so
        # that we get notifications of _new_ services.
        #if not more:
        #    browser.stop

    def netServiceBrowser_didRemoveService_moreComing_(self, browser, service, more):
        # TODO - when a services vanishes again, we should remove it from
        # the list of services, except that there doesn't seem to be any way
        # of relating the removed service to the one that got added. We may
        # have to resolve the service and remove it that way. Ick.
        #try:
        #    del self.services[ id(service) ]
        #except (KeyError):
        #    return
        #
        #if self.delegate: self.delegate.gotMoreServices(self)

        # leave the browser running
        #if not more:
        #    browser.stop
        pass
        
    def netServiceWillResolve_(self, service):
        # schedule a stop of the resolve, because (a) this is nicer to the
        # network, and (b) for a reason I don't understand, having a selector
        # scheduled against this object actually makes the thing resolve, as
        # oppsed to it just drifting off into nowhere
        service.performSelector_withObject_afterDelay_("stop", None, 5)

    def netServiceDidResolveAddress_(self, service):
        # we have managed to resolve at least one address for the service,
        # so tell the delegate so it can rebuild the list.
        self.services[ id(service) ] = service
        if self.delegate: self.delegate.gotMoreServices(self)





###################################################################
# classes representing hosts and services

class Host(NSObject):

    def init(self):
        self = NSObject.init(self)
        self.name = "name"
        self.port = ""
        self.ip = ""
        self.services = []
        return self

    def service_sort(self, first, second):
        try:
            p1 = reverse_protocols[ first.type ]
        except (KeyError):
            p1 = 9999

        try:
            p2 = reverse_protocols[ second.type ]
        except (KeyError):
            p2 = 9999

        return cmp(p1, p2)

    def addService(self, service):
        self.services.append(service)
        self.services.sort( self.service_sort )
        self.name = "%s (%s)"%(self.services[0].desc,self.ip)


class Service(NSObject):

    def init(self):
        self = NSObject.init(self)
        self.name = ""
        self.port = 0
        return self

    def service(self, service, type):

        self.type = type
        self.desc = service.name()

        # IPs resolve in various random orders. For now, pick a 'definitive'
        # one by sorting them. HACK HACK HAHCK
        self.addresses = list(service.addresses())
        self.addresses.sort( lambda a,b: cmp(a[0], b[0]) )
        
        self.ip = self.addresses[0][0]
        self.port = self.addresses[0][1]

        self.other = {}
        other = service.protocolSpecificInformation()
        if other:
            atoms = other.split("\x01")
            for a in atoms:
                try:
                    (key, value) = a.split("=")
                    self.other[key] = value
                except (ValueError):
                    pass

        try:
            protocol = protocols[ reverse_protocols[ type ] ]
            self.name = protocol['name']

            try:
                # try a custom class to customize the service display
                protocol['custom']().customize(self)
            except (KeyError):
                pass # never mind

        except (KeyError):
            # unknown type
            self.name = type

        return self

###################################################################
# custom service handling code

class ichat2(object):

    def customize(self, service):
        try:
            service.desc = "%s %s"%(service.other['1st'],service.other['last'])
        except (KeyError):
            pass


###################################################################
# all supporterd protocos

protocols = [

    # special for service discovery
    { 'type':'_services._dns-sd._udp.', 'name':'meta-query',             },

    { 'type':'_presence._tcp.',      'name':'iChat 2 presence',             'custom':ichat2 },
    { 'type':'_ichat._tcp.',         'name':'iChat 1 presence',             },
    { 'type':'_ssh._tcp.',           'name':'Remote login',                 },
    { 'type':'_afpovertcp._tcp.',    'name':'Personal file sharing',        },
    { 'type':'_hydra._tcp.',         'name':'SubEthaEdit document',         },
    { 'type':'_workstation._tcp.',   'name':'Workgroup Manager',            },
    { 'type':'_ftp._tcp.',           'name':'FTP server',                   },
    { 'type':'_distcc._tcp.',        'name':'Xcode distributed compiler',   },
    { 'type':'_iconquer._tcp.',      'name':'iConquer game server',         },
    { 'type':'_raop._tcp.',          'name':'AirTunes speaker',             },
    { 'type':'_airport._tcp.',       'name':'Airport base station',         },
    { 'type':'_printer._tcp.',       'name':'LPR printer sharing',          },
    { 'type':'_ipp._tcp.',           'name':"Internet Printing Protocol",   },
    { 'type':'_eppc._tcp.',          'name':"Remote AppleEvents",           },
    { 'type':'_smb._tcp.',           'name':'Windows file sharing',         },
    { 'type':'_clipboard._tcp.',     'name':'Shared clipboard',             },
    { 'type':'_teleport._tcp.',      'name':'Teleport server',              },
    { 'type':'_nfs._tcp.',           'name':'NFS server',                   },
    { 'type':'_omni-bookmark._tcp.', 'name':'OmniWeb shared bookmarks',     },
    { 'type':'_webdav._tcp.',        'name':'WebDav server',                },
    { 'type':'_dpap._tcp.',          'name':'iPhoto shared photos',         },
    { 'type':'_http._tcp.',          'name':'Web server',                   },
    { 'type':'_daap._tcp.',          'name':'iTunes shared music',          },
    { 'type':'_dacp._tcp.',          'name':'iTunes remote control',        },
    { 'type':'_spike._tcp.',         'name':'Spike shared clipboard',       },
    { 'type':'_beep._tcp.',          'name':'Xgrid distributed computing',  },
    { 'type':'_feed-sharing._tcp.',  'name':'NetNewsWire shared feed list', },
    { 'type':'_riousbprint._tcp.',   'name':'Airport Express printer sharing', },
]

# build a reverse-lookup of the protocol list for prioritizartion
reverse_protocols = {}
# TODO - this should be a really simple for loop, but I can't ermember how
try:
    i=0
    while 1:
        p = protocols[i]
        reverse_protocols[ p['type'] ] = i
        i = i + 1
except (IndexError):
    # end of list
    pass

if __name__ == "__main__":
    AppHelper.runEventLoop()

